from django.db import models
from django.core.urlresolvers import reverse
from django.utils.html import format_html, mark_safe
from django.forms.utils import flatatt
import json
import os
import sys
from django.core.validators import MinValueValidator


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(BASE_DIR + '/carrot')


class MessageLog(models.Model):
    """
    MessageLogs store information about a carrot task

    Lifecycle:
        #. A :class:`carrot.objects.Message` object is created and published.
        #. The act of publishing the message creates a MessageLog object with the status 'PUBLISHED'. The task now sits
           in the RabbitMQ queue until it has been consumed
        #. When a consumer digests the message, the status is updated to 'COMPLETED' if the task completes successfully
           or 'FAILED' if it encounters an exception. The output, traceback, exception message and logs are written
           back to the MessageLog object
        #. If a task has failed, it can be requeued. Requeueing a task will create a new :class:`carrot.objects.Message`
           object with the same parameters. In this case, the originally MessageLog object will be deleted
        #. If the task has been completed successfully, it will be deleted three days after completion, provided that
           the :function:`carrot.helper_tasks.cleanup` has not been disabled

    """
    STATUS_CHOICES = (
        ('UNPUBLISHED', 'Not yet published'),
        ('PUBLISHED', 'Published'),
        ('FAILED', 'Failed'),
        ('COMPLETED', 'Completed'),
    ) #:

    status = models.CharField(max_length=10, choices=STATUS_CHOICES, default='UNPUBLISHED')
    exchange = models.CharField(max_length=200, blank=True, null=True) #: the exchange
    queue = models.CharField(max_length=200, blank=True, null=True)
    routing_key = models.CharField(max_length=200, blank=True, null=True)
    uuid = models.CharField(max_length=200)
    priority = models.PositiveIntegerField(default=0)

    task = models.CharField(max_length=200)#: the import path for the task to be executed
    task_args = models.TextField(null=True, blank=True, verbose_name='Task positional arguments')
    content = models.TextField(null=True, blank=True, verbose_name='Task keyword arguments')

    exception = models.TextField(null=True, blank=True)
    traceback = models.TextField(null=True, blank=True)
    output = models.TextField(null=True, blank=True)

    publish_time = models.DateTimeField(null=True, blank=True)
    failure_time = models.DateTimeField(null=True, blank=True)
    completion_time = models.DateTimeField(null=True, blank=True)

    log = models.TextField(blank=True, null=True)

    @property
    def virtual_host(self):
        from carrot.utilities import get_host_from_name
        return get_host_from_name(self.queue)

    @property
    def keywords(self):
        """
        Used in :class:`carrot.views.MessageView` to display the keyword arguments as a table
        """
        return json.loads(self.content or '{}')

    def __str__(self):
        return self.task

    def get_url(self):
        """
        Creates a url that points to :class:`carrot.views.MessageView`
        """
        return reverse('task-info', args=[self.pk, 'INFO'])

    @property
    def href(self):
        """
        Renders self.get_url() as a HTML href
        """
        return format_html('<a {}>{}</a>', flatatt({'href': self.get_url()}), self.task)

    def display_time(self, time):
        """
        Helper function to format time for display
        """
        if time:
            return time.strftime('%Y-%m-%d %I:%M %p')

    @property
    def display_publish_time(self):
        return self.display_time(self.publish_time)

    @property
    def display_completion_time(self):
        return self.display_time(self.completion_time)

    @property
    def display_failure_time(self):
        return self.display_time(self.failure_time)

    @property
    def retry_url(self):
        """
        Returns the URL required for requeueing the task
        """
        return reverse('requeue-task', args=[self.pk])

    @property
    def delete_url(self):
        """
        Returns the URL required for requeueing the task
        """
        return reverse('delete-task', args=[self.pk])

    @property
    def positionals(self):
        import ast
        if self.task_args == '()':
            return ()
        else:
            return [ast.literal_eval(arg.strip()) for arg in self.task_args[1:-1].split(',')]

    # noinspection PyTypeChecker
    def requeue(self):
        """
        Sends a failed MessageLog back to the queue. The original MessageLog is deleted
        """
        from carrot.utilities import publish_message
        publish_message(self.task, *self.positionals, priority=self.priority, queue=self.queue, exchange=self.exchange,
                        routing_key=self.routing_key, **self.keywords)

        self.delete()

    class Meta:
        ordering = '-priority', 'pk',


class ScheduledTask(models.Model):
    """
    A model for scheduling tasks to run at a certain interval
    """

    INTERVAL_CHOICES = (
        ('seconds', 'seconds'),
        ('minutes', 'minutes'),
        ('hours', 'hours'),
        ('days', 'days'),
    )

    interval_type = models.CharField(max_length=200, choices=INTERVAL_CHOICES, default='seconds')
    interval_count = models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])

    exchange = models.CharField(max_length=200, blank=True, null=True)
    routing_key = models.CharField(max_length=200, blank=True, null=True)
    queue = models.CharField(max_length=200, blank=True, null=True)
    task = models.CharField(max_length=200)
    task_args = models.TextField(null=True, blank=True, verbose_name='Positional arguments')
    content = models.TextField(null=True, blank=True, verbose_name='Keyword arguments')

    active = models.BooleanField(default=True)

    def get_absolute_url(self):
        return reverse('edit-scheduled-task', args=[self.pk])

    @property
    def interval_display(self):
        return 'Every %i %s' % (self.interval_count, self.interval_type if self.interval_count > 1 else
            self.interval_type[:-1])

    @property
    def multiplier(self):
        if self.interval_type == 'minutes':
            return 60

        if self.interval_type == 'hours':
            return 60 * 60

        if self.interval_type == 'days':
            return 86400

        return 1

    @property
    def positional_arguments(self):
        if self.task_args:
            return tuple([a.strip() for a in self.task_args.split(',') if a])
        else:
            return ()

    def publish(self, priority=0):
        from carrot.utilities import publish_message
        kwargs = json.loads(self.content or '{}')
        if isinstance(kwargs, str):
            kwargs = {}
        return publish_message(self.task, *self.positional_arguments, priority=priority, queue=self.queue,
                               exchange=self.exchange or '', routing_key=self.routing_key or self.queue,
                               **kwargs)

    def __str__(self):
        return self.task

    @property
    def href(self):
        return format_html('<a {}>{}</a>', flatatt({'href': reverse('edit-scheduled-task', args=[self.pk])}), self.task)

    @property
    def delete_href(self):
        return format_html('<a {}>{}</a>', flatatt({'href': reverse('delete-scheduled-task', args=[self.pk])}),
                           self.task)





from django.views.generic.list import ListView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from carrot.models import MessageLog, ScheduledTask
from django.shortcuts import redirect
from django.core.urlresolvers import reverse
from django.urls import reverse_lazy
from django import forms
from django.conf import settings
import json
import ast
import importlib
from inspect import getmembers, isfunction


class TaskForm(forms.ModelForm):
    def __init__(self, **kwargs):
        super(TaskForm, self).__init__(**kwargs)
        modules = settings.CARROT.get('task_modules', None)
        if modules:
            task_choices = []
            for module in modules:
                try:
                    mod = importlib.import_module(module)
                    functions = [o[0] for o in getmembers(mod) if isfunction(o[1]) and not o[0] == 'task']

                    for function in functions:
                        f = '%s.%s' % (module, function)
                        task_choices.append((f, f))
                except (ImportError, AttributeError):
                    pass

            self.fields['task'] = forms.ChoiceField(choices=task_choices)

    def clean(self):
        cleaned_data = super(TaskForm, self).clean()

        if not self.cleaned_data['queue']:
            raise forms.ValidationError({'queue': 'This field is required'})
        try:
            json.loads(self.cleaned_data['content'] or '{}')
        except json.JSONDecodeError:
            raise forms.ValidationError({'content': 'this field must be json serializable'})

        if self.cleaned_data.get('task_args', None):
            for arg in self.cleaned_data['task_args'].split(','):
                try:
                    ast.literal_eval(arg.strip())
                except Exception as err:
                    raise forms.ValidationError({'task_args': 'This string cannot be parsed. Reason: %s' % err})

        return cleaned_data

    class Meta:
        model = ScheduledTask
        fields = (
            'task', 'task_args', 'queue', 'content', 'exchange', 'routing_key', 'interval_type', 'interval_count',
            'active'
        )


class MessageList(ListView):
    queryset = MessageLog.objects.filter(status='PUBLISHED')

    def get_context_data(self, **kwargs):
        context = super(MessageList, self).get_context_data(**kwargs)
        context['failed_tasks'] = MessageLog.objects.filter(status='FAILED')
        context['completed_tasks'] = MessageLog.objects.filter(status='COMPLETED')
        context['scheduled_tasks'] = ScheduledTask.objects.all()
        return context


def requeue(request, pk):
    msg = MessageLog.objects.get(pk=pk)
    msg.requeue()
    return redirect(reverse('carrot-monitor'))


def requeue_all(request):
    for msg in MessageLog.objects.filter(status='FAILED'):
        msg.requeue()
    return redirect(reverse('carrot-monitor'))


def delete_all(request):
    for msg in MessageLog.objects.filter(status='FAILED'):
        msg.delete()
    return redirect(reverse('carrot-monitor'))


class DeleteFailedTaskView(DeleteView):
    model = MessageLog
    pk_url_kwarg = 'pk'
    success_url = reverse_lazy('carrot-monitor')


class MessageView(DetailView):
    model = MessageLog
    pk_url_kwarg = 'pk'

    def get_context_data(self, **kwargs):
        context = super(MessageView, self).get_context_data(**kwargs)
        context['loglevel'] = self.kwargs.get('level', "INFO")
        return context


class CreateScheduledTaskView(CreateView):
    model = ScheduledTask
    form_class = TaskForm


class UpdateScheduledTaskView(UpdateView):
    model = ScheduledTask
    pk_url_kwarg = 'pk'
    form_class = TaskForm


class DeleteScheduledTaskView(DeleteView):
    model = ScheduledTask
    pk_url_kwarg = 'pk'
    success_url = reverse_lazy('carrot-monitor')








from django.conf.urls import url
from carrot.views import (
    MessageList, MessageView, requeue, CreateScheduledTaskView, UpdateScheduledTaskView, DeleteScheduledTaskView,
    DeleteFailedTaskView, requeue_all, delete_all
)
from carrot.utilities import decorate_class_view, decorate_function_view
from django.conf import settings

decorators = settings.CARROT.get('monitor_authentication', [])


def _(v, **kwargs):
    return decorate_class_view(v, decorators).as_view(**kwargs)


def _f(v):
    return decorate_function_view(v, decorators)


urlpatterns = [
    url(r'^$', _(MessageList), name='carrot-monitor'),
    url(r'^(?P<pk>[0-9]+)/view/level=(?P<level>[a-zA-Z0-9-]+)$', _(MessageView), name='task-info'),
    # url(r'^(?P<pk>[0-9]+)/view?$', _(MessageView), name='task-info'),
    url(r'^(?P<pk>[0-9]+)/requeue/$', _f(requeue), name='requeue-task'),
    url(r'requeue-all/$', _f(requeue_all), name='requeue-all'),
    url(r'delete-all/$', _f(delete_all), name='delete-all'),
    url(r'^(?P<pk>[0-9]+)/delete/$', _(DeleteFailedTaskView), name='delete-task'),
    url(r'^scheduled/(?P<pk>[0-9]+)/$', _(UpdateScheduledTaskView), name='edit-scheduled-task'),
    url(r'^scheduled/(?P<pk>[0-9]+)/delete/$', _(DeleteScheduledTaskView), name='delete-scheduled-task'),
    url(r'scheduled/create/$', _(CreateScheduledTaskView), name='create-scheduled-task'),
]


import mock
import time
import logging
import json

from django.test import TestCase
from django.core.management import call_command
from django.core.urlresolvers import reverse
from django.db.utils import OperationalError
from django.db.models import QuerySet

from carrot.models import ScheduledTask, MessageLog
from carrot.consumer import ConsumerSet, Consumer
from carrot.objects import VirtualHost, Message, DefaultMessageSerializer
from carrot.scheduler import ScheduledTaskManager, ScheduledTaskThread
from carrot.utilities import (
    get_host_from_name, publish_message, create_consumer_set, create_scheduled_task, JsonConverter, decorate_class_view,
    decorate_function_view
)
from carrot.views import (
    MessageList, requeue, TaskForm,
)
from carrot.templatetags.filters import (
    task_queue, strapline_with_url, outputblock, failed_task_queue, completed_task_queue,
    scheduled_task_queue, get_attr, table_strapline, table_strapline_completed, table_strapline_failed,
    formatted_traceback, jsonblock
)
from carrot import DEFAULT_BROKER


logger = logging.getLogger('carrot')


def test_task(*args, **kwargs):
    logger.error('blah')


def dict_task(*args, **kwargs):
    return {'blah': True}


def test_failure(*args, **kwargs):
    raise Exception('Exception from test_failure task')


ALT_CARROT = {
    'default_broker': DEFAULT_BROKER,
    'queues': [
        {
            'name': 'test',
            'host': DEFAULT_BROKER,
            'consumer_class': 'carrot.consumer.Consumer',
        },
        {
            'name': 'default',
            'host': DEFAULT_BROKER,
            'consumer_class': 'carrot.consumer.Consumer',
        }
    ],
    'task_modules': ['carrot.tests', 'carrot.invalid']
}


def mocked_model(model, count=0):
    class MockQuerySet(QuerySet):
        def count(self):
            return count

    return MockQuerySet(model=model)


def mocked_object(*args, **kwargs):
    class MockMessageLog(object):
        task = 'blah2'

        def __init__(self, *args, **kwargs):
            print(args, kwargs)
            self.task = 'blah'

        def save(*args, **kwargs):
            raise OperationalError('OperationalError')

    return MockMessageLog


def mocked_connection(**kwargs):
    class MethodFrame:
        delivery_tag = 1

        def __iter__(self):
            for i in []:
                yield

    class Properties:
        def __init__(self, fail_properties, task):
            self.fail_properties = fail_properties
            self.task = task

        @property
        def headers(self):
            if self.fail_properties:
                raise KeyError('KeyError')

            return {'type': 'carrot.tests.%s' % self.task}

        message_id = 1

    class Channel:
        def __init__(self, task, fail_properties=False):
            self.fail_properties = fail_properties
            self.task = task

        def consume(self, *args, **kwargs):
            mframe = MethodFrame()
            properties = Properties(self.fail_properties, self.task)
            content = b'{}'
            return [mframe, properties, content], properties, content

        def basic_nack(self, *args, **kwargs):
            return

        def basic_ack(self, *args, **kwargs):
            return

    class Connection(object):
        def __init__(self, *args, **kwargs):
            self.fail_properties = kwargs.pop('fail_properties', False)
            self.task = kwargs.pop('task', 'test_task')

        def sleep(self, duration):
            return

        def channel(self):
            return Channel(self.task, self.fail_properties)

    def make_connection(**options):
        return Connection(**kwargs)

    return make_connection


def mock_task(*args, **kwargs):
    class MockTask(object):
        multiplier = 1
        interval_count = 1
        pk = 1
        task = 'carrot.tests.test_task'

        def __init__(self, *args, **kwargs):
            print(args, kwargs)

        @staticmethod
        def publish():
            return

    return MockTask()


@mock.patch('carrot.objects.VirtualHost')
def consume_one(signal, *args, **kwargs):
    @mock.patch('pika.BlockingConnection', mocked_connection(**kwargs))
    def wrap(*args):
        consumer = Consumer(1, 'test', VirtualHost(), 'null', logger, run_once=True)
        consumer.signal = signal
        consumer.start()
        consumer.join()

    return wrap()


class CarrotTestCase(TestCase):
    @mock.patch('pika.BlockingConnection')
    @mock.patch('carrot.consumer.ConsumerSet')
    @mock.patch('time.sleep')
    def test_commands(self, *args):
        kwargs = {'logfile': 'mycarrot.log', 'testmode': True}

        with self.settings(CARROT=ALT_CARROT):
            with mock.patch('time.sleep'):
                with mock.patch('carrot.models.ScheduledTask.objects', mocked_model(ScheduledTask, count=300)):
                    with self.assertRaises(SystemExit):
                        call_command('carrot', **kwargs)
                with mock.patch('carrot.models.ScheduledTask.objects', mocked_model(ScheduledTask, count=-1)):
                    with self.assertRaises(SystemExit):
                        call_command('carrot', **kwargs)

        with mock.patch('carrot.scheduler.ScheduledTaskManager.start',
                        side_effect=Exception('Generic Exception from test_commands')):
            call_command('carrot', **kwargs)

    @mock.patch('carrot.objects.VirtualHost')
    @mock.patch('carrot.consumer.Consumer')
    def test_consumer_set(self, *args):
        with mock.patch('pika.BlockingConnection', mocked_connection()):
            with mock.patch('carrot.models.MessageLog.objects.get', mocked_object):
                c = ConsumerSet(VirtualHost(), 'null', logfile='null.log',)
                c.start_consuming()
                time.sleep(0.01)
                with self.assertRaises(SystemExit):
                    c.stop_consuming()

    @mock.patch('time.sleep')
    def test_consumer(self, *args):
        consume_one()

        with mock.patch('carrot.models.MessageLog.objects.get'):
            consume_one(True, task='dict_task')
            print('logging task start')
            consume_one(True)
            print('logging task end')
            consume_one(False)
            consume_one(True, task='invalid_task')
            consume_one(True, fail_properties=True)

            with mock.patch('carrot.objects.DefaultMessageSerializer.serialize_arguments', side_effect=Exception('test')):
                consume_one(task='dict_task')

        with mock.patch('carrot.models.MessageLog.objects.get', mocked_object):
            consume_one(True, task='test_task')

    @mock.patch('pika.BlockingConnection')
    def test_utilities(self, *args):
        host = get_host_from_name(None)
        self.assertTrue(isinstance(host, VirtualHost))

        with self.settings(CARROT=ALT_CARROT):
            host = get_host_from_name('test')
            self.assertTrue(isinstance(host, VirtualHost))

            with self.assertRaises(Exception):
                get_host_from_name('blah')

            c = create_consumer_set('test', logfile='null.log')
            self.assertTrue(isinstance(c, ConsumerSet))

        publish_message('carrot.tests.test_task')
        publish_message(test_task)
        with self.assertRaises(ImportError):
            publish_message('invalid.path')

        with self.assertRaises(AttributeError):
            publish_message('carrot.tests.invalid_task')

        create_scheduled_task(test_task, {'seconds': 1}, 'test')

        with self.assertRaises(AttributeError):
            create_scheduled_task(test_task, {'seconds': 1, 'hours': 1}, 'test')

        JsonConverter().convert(json.dumps({'blah': True}), first_row='blah')
        JsonConverter().convert('', first_row='blah')

        decorate_class_view(MessageList, None)
        decorate_function_view(requeue, None)

        cls_view = decorate_class_view(MessageList, ['django.contrib.auth.decorators.login_required'])
        func_view = decorate_function_view(requeue, ['django.contrib.auth.decorators.login_required'])

        from django.test import RequestFactory
        from django.contrib.auth.models import User
        r = RequestFactory().get(cls_view())
        r.user = mock.MagicMock(spec=User)

        func_view(r, pk=1)
        with self.assertRaises(AttributeError):
            print(cls_view().dispatch(r))

    @mock.patch('pika.BlockingConnection')
    def test_models(self, *args):
        from django.utils.safestring import SafeText
        msg = publish_message(test_task, 0, None)
        self.assertTrue(isinstance(msg.virtual_host, VirtualHost))
        self.assertTrue(isinstance(msg.keywords, dict))
        self.assertTrue(isinstance(msg.href, SafeText))
        self.client.get(msg.get_url())
        self.client.get(msg.retry_url)
        self.client.get(msg.delete_url)

        task = create_scheduled_task(test_task, {'seconds': 5})

        self.client.get(task.get_absolute_url())
        self.assertEqual(task.interval_display, 'Every 5 seconds')

        self.assertEqual(task.multiplier, 1)
        task.interval_type = 'minutes'
        self.assertEqual(task.multiplier, 60)
        task.interval_type = 'hours'
        self.assertEqual(task.multiplier, 3600)
        task.interval_type = 'days'
        self.assertEqual(task.multiplier, 86400)
        self.assertEqual(task.positional_arguments, ())
        task.task_args = 'blah, blah2,'
        self.assertEqual(task.positional_arguments, ('blah','blah2'))

        self.assertTrue(isinstance(task.publish(), MessageLog))
        self.assertTrue(isinstance(task.href, SafeText))
        self.assertTrue(isinstance(task.delete_href, SafeText))

    def test_objects(self):
        self.assertEqual('amqp://guest:guest@localhost:5672//', str(VirtualHost(url='amqp://localhost:5672')))
        with self.assertRaises(Exception):
            VirtualHost(url='blah')
        msg = Message('carrot.tests.test_task', VirtualHost(url=DEFAULT_BROKER), task_kwargs={'blah': True})

        serializer = DefaultMessageSerializer(msg)
        self.assertTrue(isinstance(serializer.body(), str))
        with self.settings(CARROT=ALT_CARROT):
            Message('carrot.tests.test_task', None)

    def test_scheduler(self):
        task = create_scheduled_task(test_task, {'seconds': 1}, blah=True)
        s = ScheduledTaskManager()
        s.start()
        s.stop()

        with mock.patch('carrot.utilities.publish_message'):
            with mock.patch('carrot.models.ScheduledTask.objects.get', mock_task):
                thread = ScheduledTaskThread(task, True)
                thread.active = False
                thread.inactive_reason = 'test'

                with mock.patch('time.sleep'):
                    thread.start()
                    thread.join()

                    thread = ScheduledTaskThread(task)
                    thread.start()

                thread.active = False
                thread.join()

    def test_views(self):
        log = publish_message('carrot.tests.test_task')
        log.status = 'FAILED'
        log.save()
        self.assertEqual(self.client.get(reverse('requeue-all')).status_code, 302)
        log.status = 'FAILED'
        log.save()
        self.assertEqual(self.client.get(reverse('delete-all')).status_code, 302)

        valid_data = {
            'content': '{"blah": "yes"}',
            'task': 'carrot.tests.test_task',
            'queue': 'default',
            'interval_type': 'seconds',
            'interval_count': 1,
            'task_args': '"blah"',

        }
        form = TaskForm(data=valid_data)

        self.assertTrue(form.is_valid(), form.errors)

        invalid_data = {
            'queue': 'default',
            'content': '{"2]34[]er[2][e][234][][{}{}"}',
        }
        form = TaskForm(data=invalid_data)
        self.assertFalse((form.is_valid()))

        invalid_data = {}

        with self.settings(CARROT=ALT_CARROT):
            form = TaskForm(data=invalid_data)
            self.assertFalse((form.is_valid()))

        invalid_data = {
            'queue': 'default',
            'task_args': 'blah',
        }

        with self.settings(CARROT=ALT_CARROT):
            form = TaskForm(data=invalid_data)
            self.assertFalse((form.is_valid()))

    def test_filters(self):
        task_queue([])
        strapline_with_url([])
        outputblock(json.dumps({'blah': True}))
        outputblock('fjdioju438u{}{£}!"£}"£}!{£}!"£!}{£!}')
        outputblock({})

        jsonblock(MessageLog(task_args="['blah', None]", content=json.dumps({'blah':True})))
        jsonblock(MessageLog(task_args='()', content=json.dumps({'blah': True})))
        failed_task_queue(MessageLog.objects.none())
        completed_task_queue(MessageLog.objects.none())
        scheduled_task_queue(MessageLog.objects.none())
        get_attr(VirtualHost(), 'name')

        table_strapline(mocked_model(MessageLog, 0))
        table_strapline(mocked_model(MessageLog, 10))
        table_strapline(mocked_model(MessageLog, 40))

        table_strapline_completed(mocked_model(MessageLog, 0))
        table_strapline_completed(mocked_model(MessageLog, 10))
        table_strapline_completed(mocked_model(MessageLog, 40))

        table_strapline_failed(mocked_model(MessageLog, 0))
        table_strapline_failed(mocked_model(MessageLog, 10))
        table_strapline_failed(mocked_model(MessageLog, 40))

        formatted_traceback('consumer date time WARNING:: message')
        formatted_traceback('blah')









from django import template
from django.utils.html import format_html, mark_safe, format_html_join
from django.forms.utils import flatatt
import json
from carrot.utilities import JsonConverter
from django.core.urlresolvers import reverse


register = template.Library()


@register.simple_tag
def el(tag, content=None, **opts):
    if content:
        return format_html('<{} {}>{}</{}>', tag, flatatt(opts), content, tag)
    else:
        return format_html('<{} {} />', tag, flatatt(opts))


@register.simple_tag
def ot(tag, **opts):
    return format_html('<{} {}>', tag, flatatt(opts))


@register.simple_tag
def ct(tag):
    return format_html('</{}>', tag)


@register.inclusion_tag('carrot/table.html')
def task_queue(object_list):
    return {
        'table_class': 'task-queue',
        'headers': ['Priority', 'Task', 'Arguments', 'Keyword arguments', 'Queue', 'Exchange', 'Routing key'],
        'attributes': ['priority', 'href', 'task_args', 'content', 'queue', 'exchange', 'routing_key'],
        'object_list': object_list[:30]
    }


@register.inclusion_tag('carrot/strapline.html')
def strapline_with_url(null):
    return {
        'pre_msg': 'Click',
        'url': reverse('create-scheduled-task', args=[]),
        'msg': 'here',
        'post_msg': 'to create a new scheduled task'
    }


@register.inclusion_tag('carrot/task_form.html')
def task_form(form):
    return {'form': form}


@register.filter
def jsonblock(object):
    content = object.keywords
    if not content or str(content) == '"{}"':
        return format_html('<ul {}>{}</ul>',
                           flatatt({'class': 'traceback green'}),
                           'No information provided'
                           )

    converter = JsonConverter()
    if object.positionals:
        first_row = format_html('<tr><th>{}</th><th>{}</th></tr><tr><td>Positional arguments</td><td>{}</td></tr',
                                'Field', 'Value', object.positionals)
    else:
        first_row = format_html('<tr><th>{}</th><th>{}</th></tr', 'Field', 'Value')

    return mark_safe(converter.convert(
        json=object.keywords,
        first_row=first_row,
        table_attributes=flatatt({'class': 'task-queue'}),
    ))


@register.filter
def outputblock(content):
    try:
        if not content or str(content) == '"{}"':
            return format_html('<ul {}>{}</ul>',
                               flatatt({'class': 'traceback green'}),
                               'No information provided'
                               )
        converter = JsonConverter()
        first_row = format_html('<tr><th>{}</th><th>{}</th></tr', 'Field', 'Value')

        return mark_safe(converter.convert(
            json=json.loads(content or '{}'),
            first_row=first_row,
            table_attributes=flatatt({'class': 'task-queue'}),
        ))

    except json.JSONDecodeError:
        return format_html('<ul {}>{}</ul>',
                           flatatt({'class': 'traceback green'}),
                           format_html_join('\n', '<li class="tb">{}</li>', ((line,) for line in content.split('\n'))))


@register.inclusion_tag('carrot/info_table.html')
def info_table(object):
    attributes = [
        ('Virtual host', object.virtual_host,),
        ('Queue', object.queue,),
        ('Exchange', object.exchange,),
        ('Routing key', object.routing_key),
        ('Status', object.status),
        ('Priority', object.priority),
        ('Publish time', object.display_publish_time),
        ('Completion time', object.display_completion_time),
        ('Failure time', object.display_failure_time),
    ]
    return {
        'table_class': 'task-queue',
        'headers': ['Field', 'Value'],
        'attributes': attributes,
    }


@register.inclusion_tag('carrot/table.html')
def failed_task_queue(object_list):
    object_list = object_list.order_by('-failure_time')
    return {
        'table_class': 'task-queue failed',
        'headers': ['Failure Time', 'Task', 'Arguments', 'Keyword arguments', 'Exchange', 'Routing key', 'Exception'],
        'attributes': ['display_failure_time', 'href', 'task_args', 'content', 'exchange', 'routing_key', 'exception'],
        'object_list': object_list[:30]
    }


@register.inclusion_tag('carrot/table.html')
def completed_task_queue(object_list):
    object_list = object_list.order_by('-completion_time')
    return {
        'table_class': 'task-queue completed',
        'headers': ['Completion Time', 'Task', 'Arguments', 'Keyword arguments', 'Exchange', 'Routing key',],
        'attributes': ['display_completion_time', 'href', 'task_args', 'content', 'exchange', 'routing_key',],
        'object_list': object_list[:30]
    }


@register.inclusion_tag('carrot/table.html')
def scheduled_task_queue(object_list):
    return {
        'table_class': 'task-queue',
        'headers': ['Task', 'Interval', 'Exchange', 'Routing key', 'Active?'],
        'attributes': ['href', 'interval_display', 'exchange', 'routing_key', 'active'],
        'object_list': object_list[:30]
    }



@register.simple_tag
def get_attr(instance, field, default=None):
    return getattr(instance, field, default)


@register.filter
def table_strapline(object_list):
    if object_list.count() > 30:
        msg = 'Showing the next 30 tasks to be processed (out of a total of %i)' % object_list.count()
    elif object_list.count() == 0:
        msg = 'No items currently in the queue'

    else:
        msg = 'Showing all %i tasks currently in the queue' % object_list.count()

    return format_html('<p {}>{}<p>', flatatt({'class':'strapline'}), msg)


@register.filter
def table_strapline_failed(object_list):
    requeue_delete = 'Click <a {}>here</a> to requeue all tasks, or <a {}>here</a> to delete all tasks'

    requeue_delete_args = flatatt({'href': reverse('requeue-all')}), flatatt({'href': reverse('delete-all')})

    if object_list.count() > 30:
        msg = 'Showing the 30 most recent failed tasks (out of a total of %i).' % object_list.count(),

    elif object_list.count() == 0:
        requeue_delete_args = ()
        requeue_delete = ''
        msg = 'No failed tasks'

    else:
        msg = 'Showing all %i failed tasks.' % object_list.count()

    return format_html('<p {}>{} %s<p>' % requeue_delete, flatatt({'class':'strapline'}), msg, *requeue_delete_args)


@register.filter
def table_strapline_completed(object_list):
    if object_list.count() > 30:
        msg = 'Showing the 30 most recent completed tasks (out of a total of %i)' % object_list.count()

    elif object_list.count() == 0:
        msg = 'No recent completed tasks'

    else:
        msg = 'Showing all %i completed tasks' % object_list.count()

    return format_html('<p {}>{}<p>', flatatt({'class':'strapline'}), msg)


@register.filter
def formatted_traceback(traceback, loglevel='WARNING'):
    levels = {
        'DEBUG': 0,
        'INFO': 1,
        'WARNING': 2,
        'ERROR': 3,
        'CRITICAL': 4,
    }

    items = []
    traceback = traceback.split('\n')
    for line in traceback:
        def get_indent(l):
            return int((len(l) - len(l.lstrip())) / 2)
        try:
            data, msg = line.strip().split('::')[0:2]
            consumer, date, time, level = data.split()
            msg = '%s %s: %s' % (date, time, msg)
            if levels[level] >= levels[loglevel]:
                items.append((get_indent(line), level, msg))
        except ValueError:
            items.append((get_indent(line), 'ERROR', line.strip()))

    list_items = format_html_join('\n', '<li class="tb indent_level_{} {}">{}</li>', ((item) for item in items))

    return format_html('<ul class="traceback yellow">{}</ul>', list_items)



import threading
from carrot.models import ScheduledTask
import time
from django.core.exceptions import ObjectDoesNotExist


class ScheduledTaskThread(threading.Thread):
    """
    A thread that handles a single :class:`carrot.models.ScheduledTask` object. When started, it waits for the interval
    to pass before publishing the task to the required queue

    While waiting for the task to be due for publication, the process continuously monitors the object in the Django
    project's database for changes to the interval, task, or arguments, or in case it gets deleted/marked as inactive
    and response accordingly

    :param carrot.models.ScheduledTask scheduled_task: the scheduled task to be published periodically
    :param bool run_now: whether or not to run the task before waiting for the first interval
    :param dict filters: for limiting the queryset of ScheduledTasks for montoring (defaults to `active=True`)

    """
    def __init__(self, scheduled_task, run_now=False, **filters):
        threading.Thread.__init__(self)
        self.id = scheduled_task.id
        self.queue = scheduled_task.routing_key
        self.scheduled_task = scheduled_task
        self.run_now = run_now
        self.active = True
        self.filters = filters
        self.inactive_reason = ''

    def run(self):
        interval = self.scheduled_task.multiplier * self.scheduled_task.interval_count

        count = 0
        if self.run_now:
            self.scheduled_task.publish()

        while True:
            while count < interval:
                if not self.active:
                    if self.inactive_reason:
                        print('Thread stop has been requested because of the following reason: %s.\n Stopping the '
                              'thread' % self.inactive_reason)

                    return

                try:
                    self.scheduled_task = ScheduledTask.objects.get(pk=self.scheduled_task.pk, **self.filters)
                    interval = self.scheduled_task.multiplier * self.scheduled_task.interval_count

                except ObjectDoesNotExist:
                    print('Current task has been removed from the queryset. Stopping the thread')
                    return

                time.sleep(1)
                count += 1

            print('Publishing message %s' % self.scheduled_task.task)
            self.scheduled_task.publish()
            count = 0


class ScheduledTaskManager(object):
    """
    The main scheduled task manager project. For every active :class:`carrot.models.ScheduledTask`, a
    :class:`ScheduledTaskThread` is created and started

    This object exists for the purposes of starting these threads on startup, or when a new ScheduledTask object
    gets created, and implements a .stop() method to stop all threads

    """
    def __init__(self, **options):
        self.threads = []
        self.filters = options.pop('filters', {'active': True})
        self.run_now = options.pop('run_now', False)
        self.tasks = ScheduledTask.objects.filter(**self.filters)

    def start(self):
        print('found %i scheduled tasks to run' % self.tasks.count())
        for t in self.tasks:
            print('starting thread for task %s' % t.task)
            thread = ScheduledTaskThread(t, self.run_now, **self.filters)
            thread.start()
            self.threads.append(thread)

    def add_task(self, task):
        thread = ScheduledTaskThread(task, self.run_now, **self.filters)
        thread.start()
        self.threads.append(thread)

    def stop(self):
        print('Attempting to stop %i running threads' % len(self.threads))

        for t in self.threads:
            print('Stopping thread %s' % t)
            t.active = False
            t.inactive_reason = 'A termination of service was requested'
            t.join()
            print('thread %s stopped' % t)



from carrot.utilities import publish_message

def my_task(**kwargs):
    return 'hello world'

publish_message(my_task, hello=True)


from carrot.utilities import create_scheduled_task

create_scheduled_task(my_task, {'seconds': 5}, hello=True)


from carrot.models import MessageLog
from django.utils import timezone
from datetime import timedelta


def delete_expired(expiry=None):
    """
    Finds all completed :class:`carrot.models.MessageLog` objects that were completed more than 3 days ago, and deletes
    them
    """
    if expiry is None:
        expiry = {'days': 3}

    completed_tasks = MessageLog.objects.filter(status='COMPLETED',
                                                completion_time__lte=timezone.now() - timedelta(**expiry))
    count = completed_tasks.count()
    completed_tasks.delete()

    return 'Deleted %i entries' % count



"""
This module provides a backend API for creating Consumers and Consumer Sets

"""

import json
from carrot.models import MessageLog
import traceback
import uuid
import threading
import logging
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
import importlib
import sys
from carrot.objects import DefaultMessageSerializer
from django.db.utils import OperationalError


LOGGING_FORMAT = '%(threadName)-10s %(asctime)-10s %(levelname)s:: %(message)s'


class ListHandler(logging.Handler):
    """
    A :class:`logging.Handler` that records each log entry to a python list object, provided that the entry is coming
    from the correct thread.

    Allows for task-specific logging

    """
    def __init__(self, thread_name, level):
        self.output = []
        self.thread_name = thread_name
        super(ListHandler, self).__init__(level)

    def emit(self, record):
        msg = self.format(record)
        if msg.startswith(self.thread_name):
            self.output.append(msg)

    def parse_output(self):
        return '\n'.join([str(line) for line in self.output])


class KeepAlive(threading.Thread):
    """
    This thread is started every time carrot executes a function, and sends a heartbeat to RabbitMQ so as to avoid
    dropping the connection (which can happen with long-running functions)
    """
    def __init__(self, connection):
        threading.Thread.__init__(self)
        self.connection = connection
        self.keep_alive = True

    def run(self):
        if self.keep_alive:
            self.connection.sleep(0.1)


class LoggingTask(object):
    """
    Turns a function into a class with :meth:`.run()` method, and attaches a :class:`KeepAlive` object and a
    :class:`ListHandler` logging handler
    """
    def __init__(self, task, logger, thread_name, connection, *args, **kwargs):
        self.task = task
        self.args = args
        self.kwargs = kwargs

        self.logger = logger
        self.thread_name = thread_name
        self.out = []
        self.stream_handler = ListHandler(thread_name, logger.getEffectiveLevel())
        self.stream_handler.setLevel(logger.getEffectiveLevel())
        formatter = logging.Formatter(LOGGING_FORMAT)
        self.stream_handler.setFormatter(formatter)

        self.logger.addHandler(self.stream_handler)

        self._run = False
        self._keep_alive = None
        self.connection = connection

    def keep_alive(self):
        self._ka = KeepAlive(self.connection)
        self._ka.start()

    def stop_keep_live(self):
        self._ka.keep_alive = False
        self._ka.join(1)

    def run(self):
        self.keep_alive()
        output = self.task(*self.args, **self.kwargs)
        self.stop_keep_live()
        self._run = True
        return output

    def get_logs(self):
        if self._run:
            self.logger.removeHandler(self.stream_handler)
            return self.stream_handler.parse_output()


class Consumer(threading.Thread):
    """
    An individual Consumer object. This class is run on a detached thread and watches a specific RabbitMQ queue for
    messages, and consumes them when they appear. Multiple Consumers can be linked to the same queue using a
    :class:`.ConsumerSet` object.

    This object takes the following parameters:

        :param thread_id: an identifier for the thread (a uuid)
        :param name: the name of the Consumer. Note that all Consumers in the same :class:`.ConsumerSet` are assigned
            the same name, e.g. Consumer-1, Consumer-2
        :param host: The virtual host :class:`carrot.objects.VirtualHost` object where the queue resides
        :param queue: The name of the linked queue
        :param logger: The log object from the parent process

    __init__ also specifies the following attributes:
        - **connection**: the blocking connection of the :class:`carrot.objects.VirtualHost`
        - **task_log**: a list of log entries specific to the task being executed. This list is cleared every time
          the consumer consumes a new message
        - **signal**: This is a flag that allows the consumer to run, and is initially set to **True**. When the parent
          process is killed, the parent process sets the signal to **False** which is intercepted by various break
          points in the consumer, causing the :meth:`.run` method to return

    """
    serializer = DefaultMessageSerializer()

    def __init__(self, thread_id, name, host, queue, logger, run_once=False):
        threading.Thread.__init__(self)
        self.logger = logger
        self.thread_id = thread_id
        self.name = name
        self.host = host
        self.connection = host.blocking_connection
        self.channel = self.connection.channel()

        self.queue = queue
        self.task_log = []
        self.signal = True
        self.run_once = run_once

    def get_message_log(self, properties, body):
        """
        Finds and returns the :class:`carrot.models.MessageLog` object associated with a RabbitMQ message

        By default, carrot finds this retrieving the MessageLog UUID from the RabbitMQ message properties.message_id
        attribute.

        This method can be extended by custom consumers. For example, if you are attempting to consume from a RabbitMQ
        queue containing messages that do not come from your Carrot instance, you may want to extend this method to
        create, instead of get, a MessageLog object

        :param properties: the message properties
        :param body: the message body. This is not used by default, but is included so that the function can be
                     extended in custom consumers.

        :return: a :class:`carrot.models.MessageLog` object

        .. note::
            This method does not use self.get_task_type as the intention is to get the MessageLog object before the
            **consume** method tries to do anything else. This means that if any later part of the process fails,
            the traceback and exception information can be stored to the MessageLog object for easier debugging.

        .. warning::
            If this method fails to find a matching MessageLog object, then the RabbitMQ message will be rejected.
            Depending on the configuration of your RabbitMQ queue, this may cause a loss of data. If you are
            implementing a custom consumer, then you should use
            `dead letter exchange <http://www.rabbitmq.com/dlx.html>`_ to preserve your message content

        """
        return MessageLog.objects.get(uuid=properties.message_id)

    def get_task_type(self, properties, body):
        """
        Identifies the task type, by looking up the attribute **self.task_type** in the message properties

        :param properties: the message properties
        :param body: the message body. Not used by default, but provided so that the method can be extended if necessary

        :return: The task type as a string, e.g. *myapp.mymodule.mytask*

        """
        return properties[self.serializer.type_header]

    def consume(self, method_frame, properties, body):
        """
        On finding a message in the queue, this function is called, which executes the task. The following actions
        happen in this function:
            #. The value of *self.task_log* is cleared
            #. Carrot searches for a :class:`carrot.models.MessageLog` object with a guid matching the message ID. If
               none can be found, the message is rejected. Once the MessageLog has been found, any subsequent exceptions
               can be stored against it for debugging purposes
            #. The task type is identified
            #. The python function to be executed is identified
            #. The positional and keyword arguments are identified
            #. The function is wrapped as a :class:`.LoggingTask` object and executed
            #. If the function runs successfully, the message is acknowledged. Otherwise, the message is rejected
            #. The associated :class:`carrot.models.MessageLog` object is updated with the task log, output, exception
               traceback and status (as necessary).

        ..warning:
            Note that if you have a very high number of database connections (in other words, a very high number of
            consumers or a very high number of tasks scheduled to run at the same time), then carrot may encounter a
            Database exception on saving the output of a task to the associated :class:`carrot.models.MessageLog`.
            Carrot waits 10 seconds then retries a maximum of 10 times before raising the exception, as the database
            may become unblocked once other connections are cleared. If more than 10 retries are required, Carrot will
            be unable to save the output of the task to the :class:`carrot.models.MessageLog`. If this happens, you
            should either reduce the amount of tasks you are trying to process at the same time, or edit the maximum
            number of database connections supported by your Django project's database

        """
        if self.signal:
            try:
                log = self.get_message_log(properties, body)
                self.channel.basic_ack(method_frame.delivery_tag)
            except ObjectDoesNotExist:
                self.logger.error('Unable to find a MessageLog matching the uuid %s. Ignoring this task' %
                                  properties.message_id)
                self.channel.basic_nack(method_frame.delivery_tag, requeue=False)
                return

            try:
                task_type = self.get_task_type(properties.headers, body)
            except KeyError as err:
                self.fail(log, 'Unable to identify the task type because a key was not found in the message header: %s'
                          % err)
                return

            self.logger.info('Consuming task %s, ID=%s' % (task_type, properties.message_id))

            try:
                func = self.serializer.get_task(properties, body)
            except (ValueError, ModuleNotFoundError, AttributeError) as err:
                self.fail(log, err)
                return

            try:
                args, kwargs = self.serializer.serialize_arguments(body.decode())

            except Exception as err:
                self.fail(log,
                          'Unable to process the message due to an error collecting the task arguments: %s' % err)
                return

            try:
                start_msg = '{} {} INFO:: Starting task {}.{}'.format(self.name,
                                                                 timezone.now().strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
                                                                 func.__module__, func.__name__)
                self.logger.info(start_msg)
                self.task_log = [
                    start_msg
                ]
                task = LoggingTask(func, self.logger, self.name, self.connection, *args, **kwargs)

                output = task.run()
                self.task_log.append(task.get_logs())
                self.logger.info(task.get_logs())

                success = '{} {} INFO:: Task {} completed successfully with response {}'.format(self.name,
                                                                                           timezone.now().strftime(
                                                                                               "%Y-%m-%d %H:%M:%S,%f")[
                                                                                           :-3],
                                                                                           log.task, output)
                print(success)
                self.logger.info(success)
                self.task_log.append(success)

                log.status = 'COMPLETED'
                log.completion_time = timezone.now()

                if isinstance(output, dict):
                    log.output = json.dumps(output)
                else:
                    log.output = output

                log.log = '\n'.join(self.task_log)
                save_attempts = 0

                while True:
                    save_attempts += 1

                    try:
                        log.save()
                        break
                    except OperationalError:
                        if save_attempts > 10:
                            raise OperationalError('Unable to access the database. This is probably because the number '
                                                   'of carrot threads is too high. Either reduce the amount of '
                                                   'scheduled tasks consumers, or increase the max number of '
                                                   'connections supported by your database')
                        self.connection.sleep(10)

            except Exception as err:
                self.fail(log, 'An unknown error occurred: %s' % err)

        else:
            sys.exit()

    def fail(self, log, err):
        """
        This function is called if there is any kind of error with the `.consume()` function

        :param MessageLog log: the associated MessageLog object
        :param str err: the exception

        The exception message is logged, and the MessageLog is updated with the result

        """
        self.logger.error('Task %s failed due to the following exception: %s' % (log.task, err))
        log.status = 'FAILED'
        log.failure_time = timezone.now()
        log.exception = err
        log.traceback = traceback.format_exc()
        log.save()

    def run(self):
        """
        The consume process. Waits for a message then calls :meth:`.consume' when it finds one.

        When :prop:`.signal` is set to False (which can be done by the parent :class:`.ConsumerSet` objects
        :meth:`.ConsumerSet.stop` method), this is a break point that stops the consumer from running and terminates
        the connection
        """
        self.logger.info('Started consumer %s' % self.name)
        for message in self.channel.consume(self.queue or 'default', inactivity_timeout=1):
            if not self.signal:
                self.connection.close()
                sys.exit()

            self.connection.sleep(0.1)

            if message:
                method_frame, header_frame, body = message
                self.consume(method_frame, header_frame, body)


class ConsumerSet(object):
    """
    Creates and starts a number of :class:`.Consumer` objects. All consumers must belong to the same queue

        :param host: The virtual host where the queue belongs
        :param queue: The queue name
        :param concurrency: the number of consumers to create. Defaults to 1
        :param name: the name to assign to the individual consumers. Will be rendered as *Consumer-1, Consumer-2,* etc.
        :param logfile: the path to the log file. Defaults to carrot.log
        :param loglevel: the logging level. Defaults to logging.DEBUG

    """

    @staticmethod
    def get_consumer_class(consumer_class):
        module = '.'.join(consumer_class.split('.')[:-1])
        _cls = consumer_class.split('.')[-1]
        mod = importlib.import_module(module)
        return getattr(mod, _cls)

    def __init__(self, host, queue, logger, concurrency=1, name='consumer', consumer_class='carrot.consumer.Consumer'):
        self.logger = logger
        self.host = host
        self.connection = host.blocking_connection
        self.channel = self.connection.channel()
        self.queue = queue
        self.concurrency = concurrency
        self.name = '%s-%s' % (self.queue, name)
        self.consumer_class = self.get_consumer_class(consumer_class)
        self.threads = []

    def stop_consuming(self):
        print('%i thread(s) to close' % len(self.threads))
        for t in self.threads:
            print('Closing thread %s' % t)
            t.signal = False
            t.join()
            print('Closed thread %s' % t)

    def start_consuming(self):
        """
        Creates a thread for each concurrency level, e.g. if concurrency is set to 5, 5 threads are created.

        A :class:`.Consumer` is attached to each thread and is started
        """
        for i in range(0, self.concurrency):
            thread_id = str(uuid.uuid4())
            thread = self.consumer_class(thread_id, '%s-%i' % (self.name, i + 1), self.host, self.queue, self.logger)
            self.threads.append(thread)
            thread.start()


